/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/
#include "gdamsoundspec.h"
#include <gdam/gdam-protocol.h>

static void free_mix_element(gpointer mixel, gpointer unused)
{
	GdamSoundSpecMixElement* element;
	element = (GdamSoundSpecMixElement*)mixel;
	if (element->spec)
		gdam_sound_spec_destroy(element->spec);
	g_free(element);
}
static GdamSoundSpec* mix_parse(GdamBuffer* buffer)
{
	int terminates, num_subs;
	GList* list = NULL;
	if (!gdam_buffer_parse_int(buffer, &terminates)
	 || !gdam_buffer_parse_int(buffer, &num_subs)) {
	 	g_warning("mix_parse: too short");
		return NULL;
	}
	while (num_subs-- > 0) {
		GdamSoundSpecMixElement* element;
		element = g_new(GdamSoundSpecMixElement, 1);
		list = g_list_prepend(list, element);
		element->spec = NULL;
		if (!gdam_buffer_parse_time(buffer, &element->start_time)
		 || !(element->spec = gdam_sound_spec_parse(buffer))) {
			g_warning("error parsing subsource");
			g_list_foreach(list, free_mix_element, NULL);
			g_list_free(list);
			return NULL;
		}
	}
	return gdam_sound_spec_mix_list(terminates, g_list_reverse(list));
}

static GdamSoundSpec* modulator_parse(GdamBuffer* buffer)
{
	int terminates, num_subs;
	GList* list = NULL;
	if (!gdam_buffer_parse_int(buffer, &terminates)
	 || !gdam_buffer_parse_int(buffer, &num_subs)) {
	 	g_warning("modulator_parse: too short");
		return NULL;
	}
	while (num_subs-- > 0) {
		GdamSoundSpecMixElement* element;
		element = g_new(GdamSoundSpecMixElement, 1);
		list = g_list_prepend(list, element);
		element->spec = NULL;
		if (!gdam_buffer_parse_time(buffer, &element->start_time)
		 || !(element->spec = gdam_sound_spec_parse(buffer))) {
			g_warning("error parsing subsource");
			g_list_foreach(list, free_mix_element, NULL);
			g_list_free(list);
			return NULL;
		}
	}
	return gdam_sound_spec_modulator_list(terminates, g_list_reverse(list));
}

static GdamSoundSpec* mp3_parse(GdamBuffer* buffer)
{
	char* filename;
	GdamTime start, length;
	if (!(filename=gdam_buffer_parse_string0(buffer))
	 || !gdam_buffer_parse_time(buffer, &start)
	 || !gdam_buffer_parse_time(buffer, &length)) {
	 	g_free(filename);
		return NULL;
	}
	return gdam_sound_spec_mp3(filename, start, length);
}

static GdamSoundSpec* gain_parse(GdamBuffer* buffer)
{
	int level;
	int has_sub;
	GdamSoundSpec* sub;
	if (!gdam_buffer_parse_int(buffer, &level)
	 || !gdam_buffer_parse_int(buffer, &has_sub))
	 	return NULL;
	if (!has_sub)
		return gdam_sound_spec_gain(level, NULL);
	sub = gdam_sound_spec_parse(buffer);
	if (!sub) {
		g_warning("error parsing gain's subsource");
		return NULL;
	}
	return gdam_sound_spec_gain(level, sub);
}

static GdamSoundSpec* resample_parse(GdamBuffer* buffer)
{
	int rate;
	int has_sub;
	GdamSoundSpec* sub;
	if (!gdam_buffer_parse_int(buffer, &rate)
	 || !gdam_buffer_parse_int(buffer, &has_sub))
	 	return NULL;
	if (!has_sub)
		return gdam_sound_spec_resample(rate, NULL);
	sub = gdam_sound_spec_parse(buffer);
	if (!sub) {
		g_warning("error parsing resampler's subsource");
		return NULL;
	}
	return gdam_sound_spec_resample(rate, sub);
}

static GdamSoundSpec* echo_parse(GdamBuffer* buffer)
{
	int num_echoes, reverb, i, has_sub;
	GdamSoundSpecEchoElement* els;
	GdamSoundSpec* sub;
	if (!gdam_buffer_parse_int(buffer, &reverb)
	 || !gdam_buffer_parse_int(buffer, &num_echoes))
		return NULL;
	els = (GdamSoundSpecEchoElement*)
		alloca(sizeof(GdamSoundSpecEchoElement) * num_echoes);
	for (i = 0; i < num_echoes; i++)
		if (!gdam_buffer_parse_int(buffer, &els[i].volume)
		 || !gdam_buffer_parse_time(buffer, &els[i].delay))
			return NULL;
	
	if (!gdam_buffer_parse_int(buffer, &has_sub))
	 	return NULL;
	if (!has_sub)
		sub = NULL;
	else {
		sub = gdam_sound_spec_parse(buffer);
		if (!sub) {
			g_warning("error parsing echo's subsource");
			return NULL;
		}
	}
	return gdam_sound_spec_echo_v(sub, reverb, num_echoes, els);
}
static GdamSoundSpec* sequencer_parse(GdamBuffer* buffer)
{
	int seq_size, num_subs;
	int i;
	int* vals;
	GdamSoundSpecSequencerElement* element;
	GdamSoundSpec* spec;
	if (!gdam_buffer_parse_int(buffer, &seq_size)
	 || !gdam_buffer_parse_int(buffer, &num_subs)) {
	 	return NULL;
	}
	vals = (int*) alloca(sizeof(int) * 6 * num_subs);
	for (i = 0; i < num_subs * 6; i++) {
		if (!gdam_buffer_parse_int(buffer, vals + i))
			return NULL;
	}
	element = g_new(GdamSoundSpecSequencerElement, num_subs);
	for (i = 0; i < num_subs; i++) {
		element[i].buffer_id = vals[i*6+0];
		element[i].start_offset = vals[i*6+1];
		element[i].resample_b16 = vals[i*6+2];
		element[i].gain_b16 = vals[i*6+3];
		element[i].sample_offset = vals[i*6+4];
		element[i].output_length = vals[i*6+5];
	}
	spec = gdam_sound_spec_sequencer(seq_size, num_subs, element);
	g_free(element);
	return spec;
}

static GdamSoundSpec* by_id_parse(GdamBuffer* buffer)
{
	int source_id;
	if (!gdam_buffer_parse_int(buffer, &source_id))
	 	return NULL;
	return gdam_sound_spec_by_id(source_id);
}

static GdamSoundSpec* wav_parse(GdamBuffer* buffer)
{
	char* filename;
	if (!(filename=gdam_buffer_parse_string0(buffer)))
		return NULL;
	return gdam_sound_spec_wav(filename);
}

static GdamSoundSpec* cache_parse(GdamBuffer* buffer)
{
	int buffer_id;
	gboolean loop;
	int position;
	if (!gdam_buffer_parse_int(buffer, &buffer_id)
	 && !gdam_buffer_parse_int(buffer, &loop)
	 && !gdam_buffer_parse_int(buffer, &position))
	 	return NULL;
	return gdam_sound_spec_cache(buffer_id, loop, position);
}

static GdamSoundSpec* wavetable_parse(GdamBuffer* buffer)
{
	int buffer_id;
	int hi,lo;
	gint64 hertz;
	if (!gdam_buffer_parse_int(buffer, &buffer_id)
	 && !gdam_buffer_parse_int(buffer, &hi)
	 && !gdam_buffer_parse_int(buffer, &lo))
	 	return NULL;
	hertz = ((gint64)hi << 32) | ((gint64)(guint32)lo);
	return gdam_sound_spec_wavetable(buffer_id, hertz);
}

GdamSoundSpec* gdam_sound_spec_parse   (GdamBuffer*   buffer)
{
	int type;
	if (!gdam_buffer_parse_int(buffer, &type)) {
		g_warning("gdam_sound_spec_parse: couldn't even get type");
		return NULL;
	}
	switch (type) {
		case GDAM_SOURCE_MIX: return mix_parse(buffer);
		case GDAM_SOURCE_MODULATOR: return modulator_parse(buffer);
		case GDAM_SOURCE_ECHO: return echo_parse(buffer);
		case GDAM_SOURCE_GAIN: return gain_parse(buffer);
		case GDAM_SOURCE_RESAMPLE: return resample_parse(buffer);
		case GDAM_SOURCE_MP3: return mp3_parse(buffer);
		case GDAM_SOURCE_WAV: return wav_parse(buffer);
		case GDAM_SOURCE_SEQUENCER: return sequencer_parse(buffer);
		case GDAM_SOURCE_CACHE: return cache_parse(buffer);
		case GDAM_SOURCE_WAVETABLE: return wavetable_parse(buffer);
		case GDAM_SOURCE_BY_ID: return by_id_parse(buffer);
		default: {
			g_warning("no known source of id %d", type);
			return NULL;
		}
	}
}
