/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/
#include "gdamserver.h"
#include "gdammain.h"
#include "gdamauthenticator.h"
#include <gdam/gdamdebug.h>
#include <gdam/gdamarg.h>
#include <string.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <glib.h>
#include <stdlib.h>
#include <signal.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <errno.h>
#include "gdamconfigfile.h"

#define DEFAULT_PORT 6003

static void
usage(const char* prog_name)
{
	fprintf(stderr, "usage: %s --port port_no --config config.file\n\n"
			"Run a GDAM server (on the specified port).\n",
			prog_name);
	exit(1);
}

static int bind_tcp(int port_number)
{
	struct sockaddr_in sin;
	int rv;
	int one = 1;
	bzero(&sin, sizeof(sin));
	sin.sin_port = htons(port_number);
	rv = socket(PF_INET, SOCK_STREAM, 0);
	if (rv < 0) {
		perror("error creating listening socket??");
		return -1;
	}
	fcntl(rv, F_SETFD, FD_CLOEXEC);
	if (setsockopt(rv, SOL_SOCKET, SO_REUSEADDR, &one, sizeof(one)) < 0)
		perror("warning: couldn't make sockaddr reuseable");
	if (gdam_bind_localhost) {
		struct hostent* hent;
		hent = gethostbyname("localhost");
		memcpy(&sin.sin_addr, hent->h_addr_list[0], hent->h_length);
	}
	if (bind(rv, (struct sockaddr*)(&sin), sizeof(struct sockaddr)) < 0) {
		fprintf(stderr, "bind: failed to bind to port %d: %s.\n\n",
				port_number, strerror(errno));
		if (port_number < 1024 && getuid()) {
			fprintf(stderr, "A likely reason is that only root "
					"may bind to low-numbered ports.\n");
		}
		close(rv);
		return -1;
	}
	if (listen(rv, SOMAXCONN) < 0) {
		fprintf(stderr, "error listening: %s\n", strerror(errno));
		close(rv);
		return -1;
	}
	return rv;
}

static gboolean make_server(int port_no)
{
	GdamServer* server;
	int fd = bind_tcp(port_no);
	if (fd < 0) {
			fprintf(stderr, "error binding to port %d\n",
				port_no);
			return FALSE;
		}
	server = gdam_server_new(fd,gdam_authenticator_root());
	if (!server) {
		fprintf(stderr, "error creating server\n");
		return FALSE;
	}
	gdam_server_attach(server);
	return TRUE;
}


int main(int argc, char** argv)
{
	int i;
	gboolean made_server = FALSE;
	if (argv[1] != NULL) {
		if (strcmp(argv[1], "-v") == 0
		 || strcmp(argv[1], "--version") == 0) {
		 	printf(
			     "gdam-server, version %s.\n"
		             "(C) Copyright Dave Benson, Geoff Matters. 2000.\n"
			     "\n"
			     "THIS PROGRAM COMES WITH NO WARRANTY! "
			     "USE AT YOUR OWN RISK.\n",
			     VERSION);
			exit(0);
		}
	}
	signal(SIGPIPE, SIG_IGN);
	gdam_arg_init();
	gdam_init(&argc, &argv);
	for (i = 1; i < argc; i++) {
		if (strcmp(argv[i], "--help") == 0
		 || strcmp(argv[i], "-help") == 0
		 || strcmp(argv[i], "-h") == 0) {
			usage(argv[0]);
		}
		if (strcmp(argv[i], "--port") == 0) {
			const char* port_str = argv[++i];
			int port_no;
			if (!port_str) {
				fprintf(stderr, "need argument after "
						"--port.\n");
				usage(argv[0]);
			}
			port_no = atoi(port_str);
			if (!make_server(port_no))
				exit(1);
			made_server = TRUE;
			continue;
		}
		fprintf(stderr, "unexpected option: %s\n", argv[i]);
		usage(argv[0]);
	}
	if (!made_server) {
		fprintf(stderr, "no port specified, using default %d.\n",
				DEFAULT_PORT);
		if (!make_server(DEFAULT_PORT)) {
			fprintf(stderr, "no server created... aborting...\n");
			exit(1);
		}
	}

	gdam_main();

	return 0;
}
